# Program to generate Ramachandran plot for a PDB file. 
# http://cib.cf.ocha.ac.jp/bitool/DIHED2/
# http://www.math.fsu.edu/~quine/MB_10/6_torsion.pdf
#
# Author: Dr. Walter Filgueira de Azevedo Jr.
# Date: February, 2nd 2016.

def read_PDB():
    """Function to read a PDB file"""
    # Import library
    import sys
    
    # Set up empty list
    my_atoms = []
    
    # Call read_file_name() function 
    input_file_name = read_file_name()
    
    # Try to open PDB file
    try:
        # Open PDB file
        my_fo = open(input_file_name,"r")
    except IOError:
        print("Error! I can't find file ",input_file_name)
        sys.exit("Finishing program execution!")
    
    # Looping through PDB file content
    for line in my_fo:
        if line[0:6] == "ATOM  " or line[0:6] == "HETATM":
            my_atoms.append(line)
    
    # Close file
    my_fo.close()
    
    # Return list
    return my_atoms

def read_file_name():
    """Function to read file name"""
    
    # Read input file name
    my_file = input("Type input file name => ")
    
    # Return input file name
    return my_file

def torsion(p1,p2,p3,p4):
    """Function to calculate torsion angle for atoms a,b,c, and d"""
    # Equation from http://www.stem2.org/je/proteina.pdf
    
    # Import libraries
    import numpy as np
    import math
       
    # Get coordinates for vectors q1, q2 and q3
    q1 = np.subtract(p2,p1) # b - a
    q2 = np.subtract(p3,p2) # c - b
    q3 = np.subtract(p4,p3) # d - c

    # Calculate cross vectors
    q1_x_q2 = np.cross(q1,q2)
    q2_x_q3 = np.cross(q2,q3)    
    n1 = q1_x_q2/np.sqrt(np.dot(q1_x_q2,q1_x_q2))
    n2 = q2_x_q3/np.sqrt(np.dot(q2_x_q3,q2_x_q3))

    # Calculate unit vectors
    u1 = n2
    u3 = q2/(np.sqrt(np.dot(q2,q2)))
    u2 = np.cross(u3,u1)

    # Calculate cosine and sine
    cos_theta = np.dot(n1,u1)
    sin_theta = np.dot(n1,u2)
    
    # Calculate torsion angle
    theta = -math.atan2(sin_theta,cos_theta)    # it is different from atan2 from fortran math.atan2(y,x)
    theta_deg = np.degrees(theta)
    
    # Show results
    print("theta (rad) = ",theta)
    print("theta (deg) = ",theta_deg)
    
    # Return torsion angle in degrees
    return theta_deg
    
def calc_torsion(list_of_atoms_in):
    """Function to calculate torsion angles"""
    # Import library
    import numpy as np
  
    # We have the number of rows and number of columns equal to 3
    # matrix = [[0]*column for i in range(row)]
    columns = 3
    rows = 9999
    co   = np.array([[0]*columns]*rows,float) 
    n   = np.array([[0]*columns]*rows,float) 
    ca  = np.array([[0]*columns]*rows,float) 
    
    # Set up counts
    count_res = 0
    former_res = -9999
    new_res = -9998
    found_co = 0
    found_n = 0
    found_ca = 0
    
    # Looping through list_of_atoms_in
    for line in list_of_atoms_in:
        if line[0:6] == "ATOM  " and line[13:15] == "C ":
            co[count_res,0] = float(line[30:38])
            co[count_res,1] = float(line[38:46])
            co[count_res,2] = float(line[46:54])
            new_res = int(line[22:26])
            found_co = 1
        elif line[0:6] == "ATOM  " and line[13:15] == "N ":
            n[count_res,0] = float(line[30:38])
            n[count_res,1] = float(line[38:46])
            n[count_res,2] = float(line[46:54])
            new_res = int(line[22:26])
            found_n = 1
        elif line[0:6] == "ATOM  " and line[13:15] == "CA":
            ca[count_res,0] = float(line[30:38])
            ca[count_res,1] = float(line[38:46])
            ca[count_res,2] = float(line[46:54])
            found_ca =1
            new_res = int(line[22:26])
        complete_mc = found_co*found_n*found_ca
        if former_res != new_res and complete_mc:
            former_res = new_res 
            count_res += 1
            found_co = 0
            found_n = 0
            found_ca = 0
        
    # Set up arrays for phi and psi angles 
    phi = np.zeros(count_res-1)
    psi = np.zeros(count_res-1)     
    
    # Looping through phi psi atoms in each residue
    for i in range(count_res-1):    
        # Calls torsion function        
        phi[i] = torsion(co[i],n[i+1],ca[i+1],co[i+1])
        psi[i] = torsion(n[i+1],ca[i+1],co[i+1],n[i+2])
    
    # Show torsion angles
    print("Atomic coordinates for atoms N, CA and C\n")
    print("Residue\tN[  x\t   y        z\t ]     CA[   x\t   y        z\t ]      C[  x\t   y        z\t ]")
    for i in range(count_res-1):
        print(i+1,"\t",n[i],"\t",ca[i],"\t",co[i])
    
    # Return arrays
    return phi,psi

def plot_2_arrays(x,y,x_label_in,y_label_in):
    """Function to plot two one-dimensional arrays"""
    # Import libraries
    import matplotlib.pyplot as plt
    import numpy as np
    
    # Generate plot
    plt.plot(x,y,".")       
    plt.xlim(-180,180)      # Sets axis limits
    plt.ylim(-180,180)      # Sets axis limits
    
    plt.xticks(np.arange(-180.1,180.1,30))  # Sets ticks markers for x axis
    plt.yticks(np.arange(-180.1,180.1,30))  # Sets ticks makers for y axis

    plt.xlabel(x_label_in)    # Adds axis label
    plt.ylabel(y_label_in)    # Adds axis label
    
    plt.arrow(-180,0,360,0)   # Creates an arrow
    plt.arrow(0,-180,0,360)   # Creates an arrow

    # Show plot
    plt.show()
    
    # Create a file for plot
    fig = plt.gcf()                     # Creates a figure
    fig.set_size_inches(7.0, 7.0)       # Changes figure size
    fig.savefig('rama.png', dpi=300)    # Saves figures
              
# Main function
def main():
    
    # Call function to read PDB
    my_list_of_atoms = read_PDB()

    # Call function to calculate torsion angles
    phi,psi = calc_torsion(my_list_of_atoms)

    # Call function to plot torsion angles
    plot_2_arrays(phi,psi,"Phi(deg)","Psi(deg)")

# Main program
main()